/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 */

package com.osbcp.css;

import java.io.File;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import com.osbcp.css.constants.BackgroundAttachment;
import com.osbcp.css.constants.BackgroundPosition;
import com.osbcp.css.constants.BackgroundRepeat;
import com.osbcp.css.constants.BorderStyle;
import com.osbcp.css.constants.BorderWidth;
import com.osbcp.css.constants.Clear;
import com.osbcp.css.constants.Cursor;
import com.osbcp.css.constants.Display;
import com.osbcp.css.constants.Float;
import com.osbcp.css.constants.FontStyle;
import com.osbcp.css.constants.FontVariant;
import com.osbcp.css.constants.FontWeight;
import com.osbcp.css.constants.ListStylePosition;
import com.osbcp.css.constants.ListStyleType;
import com.osbcp.css.constants.OutlineStyle;
import com.osbcp.css.constants.OutlineWidth;
import com.osbcp.css.constants.Overflow;
import com.osbcp.css.constants.Position;
import com.osbcp.css.constants.PseudoSelector;
import com.osbcp.css.constants.TextAlign;
import com.osbcp.css.constants.TextDecoration;
import com.osbcp.css.constants.TextDirection;
import com.osbcp.css.constants.TextTransform;
import com.osbcp.css.constants.VerticalAlign;
import com.osbcp.css.constants.Visibility;
import com.osbcp.css.constants.WhiteSpace;
import com.osbcp.css.unit.AUTO;
import com.osbcp.css.unit.CM;
import com.osbcp.css.unit.EM;
import com.osbcp.css.unit.EX;
import com.osbcp.css.unit.IN;
import com.osbcp.css.unit.MM;
import com.osbcp.css.unit.PC;
import com.osbcp.css.unit.PCT;
import com.osbcp.css.unit.PT;
import com.osbcp.css.unit.PX;
import com.osbcp.css.unit.Unit;

/**
 * Java representation of CSS which can be exported to a stylesheet file.
 * 
 * @author <a href="mailto:christoffer@christoffer.me">Christoffer Pettersson</a>
 */

public class CSS implements Serializable {

	private static final long serialVersionUID = 1L;

	private Map<String, Serializable> values;
	private AttributeSelector attributeSelector;
	private Map<PseudoSelector, CSS> pseudoSelectors;

	/*
	 * ************************************************************************************************************************* *
	 *                                                                                                                           *
	 *                                                        Constructors                                                       *
	 *                                                                                                                           *
	 * ************************************************************************************************************************* *
	 */

	/**
	 * Constructor that creates a new CSS instance.
	 */

	public CSS() {
		this(new HashMap<String, Serializable>(), null, new HashMap<PseudoSelector, CSS>());
	}

	/**
	 * Constructor that creates a new CSS instance with a given sheet with attributes.
	 * 
	 * @param values A map of property values selectors that should be inserted.
	 * @param attributeSelector If any attribute selector should be used.
	 * @see <a href="http://www.w3schools.com/css/css_attribute_selectors.asp">CSS Attribute Selectors</a>
	 * @param pseudoSelectors A map of pseudo selectors that should be inserted.
	 */

	private CSS(final Map<String, Serializable> values, final AttributeSelector attributeSelector, final Map<PseudoSelector, CSS> pseudoSelectors) {
		this.values = values;
		this.attributeSelector = attributeSelector;
		this.pseudoSelectors = pseudoSelectors;
	}

	/*
	 * ************************************************************************************************************************* *
	 *                                                                                                                           *
	 *                                                        Core methods                                                       *
	 *                                                                                                                           *
	 * ************************************************************************************************************************* *
	 */

	/**
	 * Returns the class name for the CSS, based on the declared attributes. This class name is also used when creating a CSS block.
	 * 
	 * @return The class name for the CSS, based on the declared attributes.
	 */

	public String getClassName() {
		return "c" + StringUtil.md5(toString());
	}

	/**
	 * Copies the current CSS object to a new CSS object.
	 * 
	 * @return The current CSS object to a new CSS object.
	 */

	public CSS copy() {
		Map<String, Serializable> newSheet = new HashMap<String, Serializable>(values);
		Map<PseudoSelector, CSS> newPseudoSelectors = new HashMap<PseudoSelector, CSS>(pseudoSelectors);
		return new CSS(newSheet, attributeSelector, newPseudoSelectors);
	}

	/**
	 * Clears a specific CSS attribute.
	 * 
	 * @param key The CSS attribute name to be cleared
	 * @return The current CSS object
	 */

	public CSS clear(final String key) {
		values.remove(key);
		return this;
	}

	/**
	 * Returns all the current declared CSS attributes.
	 * 
	 * @return All the current declared CSS attributes.
	 */

	public Map<String, Serializable> getSheet() {
		return values;
	}

	/**
	 * Sets the attribute selector.
	 * 
	 * @param attribute The attribute.
	 * @see <a href="http://www.w3schools.com/css/css_attribute_selectors.asp">CSS Attribute Selectors</a>
	 */

	public void setAttributeSelector(final Serializable attribute) {
		setAttributeSelector(attribute, null);
	}

	/**
	 * Sets the attribute selector.
	 * 
	 * @param attribute The attribute.
	 * @param value The value.
	 * @see <a href="http://www.w3schools.com/css/css_attribute_selectors.asp">CSS Attribute Selectors</a>
	 */

	public void setAttributeSelector(final Serializable attribute, final Serializable value) {
		attributeSelector = new AttributeSelector(attribute, value);
	}

	/**
	 * Exports all the CSS declared in a specific class to a specific file.
	 * 
	 * @param stylesheetClass The class that contains the declared CSS fields.
	 * @param stylesheetFile The output file to be created and written to.
	 * @param includeComments True if comments should be written as well.
	 * @throws Exception If any error occurs
	 */

	public static void export(final Class<?> stylesheetClass, final File stylesheetFile, boolean includeComments) throws Exception {
		Exporter.export(stylesheetClass, stylesheetFile, includeComments);
	}

	/**
	 * Adds a property and value.
	 * 
	 * @param property Property to add.
	 * @param value Property value to add.
	 * @return The current CSS object.
	 */

	private CSS put(final String property, final Serializable value) {
		values.put(property, value);
		return this;
	}

	/**
	 * Returns the CSS properties in a structured block, for example:
	 * 
	 * <pre>
	 * #4cafc7ba191a3557d9fe40b67957dbaa {
	 * 	background-color: color;
	 * } 
	 * </pre>
	 * 
	 * @return The CSS in a structured block
	 */

	public String toStringBlock() {

		String attributeSelectorString = attributeSelector == null ? "" : attributeSelector.toString();

		StringBuilder output = new StringBuilder("");

		output.append("." + this.getClassName() + attributeSelectorString + " {\n");

		for (final Map.Entry<String, Serializable> entry : values.entrySet()) {
			output.append("	" + entry.getKey() + ": " + entry.getValue().toString() + ";\n");
		}

		output.append("}\n\n");

		/*
		 * Print out any attached psuedo selectors
		 */

		for (final Map.Entry<PseudoSelector, CSS> entry : pseudoSelectors.entrySet()) {

			PseudoSelector pseudoSelector = entry.getKey();
			CSS psuedoCSS = entry.getValue();
			String attributeSelectorStringInner = attributeSelector == null ? "" : attributeSelector.toString();

			output.append("." + this.getClassName() + ":" + pseudoSelector + attributeSelectorStringInner + " {\n");

			for (final Map.Entry<String, Serializable> pseudoEntry : psuedoCSS.getSheet().entrySet()) {
				output.append("	" + pseudoEntry.getKey() + ": " + pseudoEntry.getValue() + ";\n");
			}

			output.append("}\n\n");

		}

		return output.toString();
	}

	/**
	 * Returns the CSS properties on a single row, for example:
	 * 
	 * <pre>
	 * background-color: color; text-align: right; color: blue;
	 * </pre>
	 * 
	 * @return The CSS properties on a single row, for example:
	 */

	public String toString() {

		StringBuilder output = new StringBuilder("");

		for (final Map.Entry<String, Serializable> entry : values.entrySet()) {
			output.append(entry.getKey() + ": " + entry.getValue().toString() + "; ");
		}

		return output.toString().trim();
	}

	/*
	 * ************************************************************************************************************************* *
	 *                                                                                                                           *
	 *                                                       Unit methods                                                        *
	 *                                                                                                                           *
	 * ************************************************************************************************************************* *
	 */

	/**
	 * Convenient variable that represents the AUTO unit
	 */

	public static AUTO AUTO = new AUTO();

	/**
	 * Creates a new centimeter (cm) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new centimeter (cm) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static CM CM(final int length) {
		return new CM(length);
	}

	/**
	 * Creates a new em CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new em CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static EM EM(final int length) {
		return new EM(length);
	}

	/**
	 * Creates a new ex CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new ex CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static EX EX(final int length) {
		return new EX(length);
	}

	/**
	 * Creates a new inch (in) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new inch (in) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static IN IN(final int length) {
		return new IN(length);
	}

	/**
	 * Creates a new millimeter (mm) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new millimeter (mm) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static MM MM(final int length) {
		return new MM(length);
	}

	/**
	 * Creates a new pica (pc) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new pica (pc) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static PC PC(final int length) {
		return new PC(length);
	}

	/**
	 * Creates a new percent (%) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new percent (%) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static PCT PCT(final int length) {
		return new PCT(length);
	}

	/**
	 * Creates a new point (pt) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new point (pt) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static PT PT(final int length) {
		return new PT(length);
	}

	/**
	 * Creates a new pixel (px) CSS unit.
	 * 
	 * @param length The numeric length value.
	 * @return A new pixel (px) CSS unit.
	 * @see <a href="http://w3schools.com/cssref/css_units.asp">CSS Units</a>
	 */

	public static PX PX(final int length) {
		return new PX(length);
	}

	/*
	 * ************************************************************************************************************************* *
	 *                                                                                                                           *
	 *                                                     Attribute methods                                                     *
	 *                                                                                                                           *
	 * ************************************************************************************************************************* *
	 */

	/*
	 * 
	 * Pseudo selectors
	 * 
	 */

	/**
	 * Adds a pseudo selector to the current CSS.
	 * 
	 * @param pseudoSelector The pseudo selector the CSS should add.
	 * @param css The CSS associated with the pseudo selector. 
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_pseudo_classes.asp">CSS Pseudo-classes</a>
	 */

	public CSS addPseudoSelector(final PseudoSelector pseudoSelector, final CSS css) {
		this.pseudoSelectors.put(pseudoSelector, css);
		return this;
	}

	/*
	 * 
	 * Background
	 * 
	 */

	/**
	 * Sets the <i>background</i> property.
	 * 
	 * @param color The color value to be set.
	 * @param url The url value to be set.
	 * @param repeat The repeat value to be set.
	 * @param position The position value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_background.asp">CSS Background</a>
	 */

	public CSS setBackground(final Serializable color, final Serializable url, final BackgroundRepeat repeat, final BackgroundPosition position) {
		return put("background", color + " url(" + url + ") " + repeat + " " + position);
	}

	/**
	 * Sets the <i>background-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-color.asp">CSS background-color Property</a>
	 */

	public CSS setBackgroundColor(final Serializable color) {
		return put("background-color", color);
	}

	/**
	 * Sets the <i>background-image</i> property.
	 * 
	 * @param url The url value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-image.asp">CSS background-image Property</a>
	 */

	public CSS setBackgroundImage(final Serializable url) {
		return put("background-image", "url(" + url + ")");
	}

	/**
	 * Sets the <i>background-repeat</i> property.
	 * 
	 * @param repeat The repeat value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-repeat.asp">CSS background-repeat Property</a>
	 */

	public CSS setBackgroundRepeat(final BackgroundRepeat repeat) {
		return put("background-repeat", repeat);
	}

	/**
	 * Sets the <i>background-attachment</i> property.
	 * 
	 * @param attachment The attachment value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-attachment.asp">CSS background-attachment Property</a>
	 */

	public CSS setBackgroundAttachment(final BackgroundAttachment attachment) {
		return put("background-attachment", attachment);
	}

	/**
	 * Sets the <i>background-position</i> property.
	 * 
	 * @param position The position value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-position.asp">CSS background-position Property</a>
	 */

	public CSS setBackgroundPosition(final BackgroundPosition position) {
		return put("background-position", position);
	}

	/**
	 * Sets the <i>background-position</i> property.
	 * 
	 * @param xpos The horizontal position.
	 * @param ypos The vertical position.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_background-position.asp">CSS background-position Property</a>
	 */

	public CSS setBackgroundPosition(final Unit xpos, final Unit ypos) {
		return put("background-position", xpos + " " + ypos);
	}

	/*
	 * Text
	 */

	/**
	 * Sets the <i>color</i> property.
	 *  
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/css_colors.asp">CSS Colors</a>
	 */

	public CSS setColor(final Serializable color) {
		return put("color", color);
	}

	/**
	 * Sets the <i>text-align</i> property.
	 * 
	 * @param alignment The alignment value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_text-align.asp">CSS text-align Property</a>
	 */

	public CSS setTextAlign(final TextAlign alignment) {
		return put("text-align", alignment);
	}

	/**
	 * Sets the <i>direction</i> property.
	 * 
	 * @param direction The direction value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_direction.asp">CSS direction Property</a>
	 */

	public CSS setDirection(final TextDirection direction) {
		return put("direction", direction);
	}

	/**
	 * Sets the <i>letter-spacing</i> property.
	 * 
	 * @param spacing The spacing value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_letter-spacing.asp">CSS letter-spacing Property</a>
	 */

	public CSS setLetterSpacing(final Unit spacing) {
		return put("letter-spacing", spacing);
	}

	/**
	 * Sets the <i>line-height</i> property.
	 * 
	 * @param height The height value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_line-height.asp">CSS line-height Property</a>
	 */

	public CSS setLineHeight(final Unit height) {
		return put("line-height", height);
	}

	/**
	 * Sets the <i>text-decoration</i> property.
	 * 
	 * @param decoration The decoration value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_text-decoration.asp">CSS text-decoration Property</a>
	 */

	public CSS setTextDecoration(final TextDecoration decoration) {
		return put("text-decoration", decoration);
	}

	/**
	 * Sets the <i>text-indent</i> property.
	 * 
	 * @param indent The indent value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_text-indent.asp">CSS text-indent Property</a>
	 */

	public CSS setTextIndent(final Unit indent) {
		return put("text-indent", indent);
	}

	/**
	 * Sets the <i>text-shadow</i> property.
	 * 
	 * @param hShadow The horizontal shadow value to be set.
	 * @param vShadow The vertical shadow value to be set.
	 * @param blur The blue value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/css3_pr_text-shadow.asp">CSS text-shadow Property</a>
	 */

	public CSS setTextShadow(final Unit hShadow, final Unit vShadow, final Unit blur, final Serializable color) {
		return put("text-shadow", hShadow + " " + vShadow + " " + blur + " " + color);
	}

	/**
	 * Sets the <i>text-transform</i> property.
	 * 
	 * @param transform The transform value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_text-transform.asp">CSS text-transform Property</a>
	 */

	public CSS setTextTransform(final TextTransform transform) {
		return put("text-transform", transform);
	}

	/**
	 * Sets the <i>vertical-align</i> property.
	 * 
	 * @param alignment The alignment value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_vertical-align.asp">CSS vertical-align Property</a>
	 */

	public CSS setVerticalAlign(final VerticalAlign alignment) {
		return put("vertical-align", alignment);
	}

	/**
	 * Sets the <i>vertical-align</i> property.
	 * 
	 * @param alignment The alignment value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_vertical-align.asp">CSS vertical-align Property</a>
	 */

	public CSS setVerticalAlign(final Unit alignment) {
		return put("vertical-align", alignment);
	}

	/**
	 * Sets the <i>white-space</i> property.
	 * 
	 * @param space The space value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_white-space.asp">CSS white-space Property</a>
	 */

	public CSS setWhiteSpace(final WhiteSpace space) {
		return put("white-space", space);
	}

	/**
	 * Sets the <i>word-spacing</i> property.
	 * 
	 * @param spacing The spacing value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_text_word-spacing.asp">CSS word-spacing Property</a>
	 */

	public CSS setWordSpacing(final Unit spacing) {
		return put("word-spacing", spacing);
	}

	/*
	 * Font
	 */

	/**
	 * Sets the <i>font-family</i> property.
	 * 
	 * @param family The family value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_font_font-family.asp">CSS font-family Property</a>
	 */

	public CSS setFontFamily(final Serializable family) {
		return put("font-family", family);
	}

	/**
	 * Sets the <i>font-size</i> property.
	 * 
	 * @param size The family value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_font_font-size.asp">CSS font-size Property</a>
	 */

	public CSS setFontSize(final Unit size) {
		return put("font-size", size);
	}

	/**
	 * Sets the <i>font-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_font_font-style.asp">CSS font-style Property</a>
	 */

	public CSS setFontStyle(final FontStyle style) {
		return put("font-style", style);
	}

	/**
	 * Sets the <i>font-variant</i> property.
	 * 
	 * @param variant The variant value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_font_font-variant.asp">CSS font-variant Property</a>
	 */

	public CSS setFontVariant(final FontVariant variant) {
		return put("font-variant", variant);
	}

	/**
	 * Sets the <i>font-weight</i> property.
	 * 
	 * @param weight The weight value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_font_font-weight.asp">CSS font-weight Property</a>
	 */

	public CSS setFontWeight(final FontWeight weight) {
		return put("font-weight", weight);
	}

	/*
	 * List 
	 */

	/**
	 * Sets the <i>list-style-type</i> property.
	 * 
	 * @param type The type value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_list-style-type.asp">CSS list-style-type Property</a>
	 */

	public CSS setListStyleType(final ListStyleType type) {
		return put("list-style-type", type);
	}

	/**
	 * Sets the <i>list-style-position</i> property.
	 * 
	 * @param position The position value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_list-style-position.asp">CSS list-style-position Property</a>
	 */

	public CSS setListStylePosition(final ListStylePosition position) {
		return put("list-style-position", position);
	}

	/**
	 * Sets the <i>list-style-image</i> property.
	 * 
	 * @param url The url value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_list-style-image.asp">CSS list-style-image Property</a>
	 */

	public CSS setListStyleImage(final Serializable url) {
		return put("list-style-image", url);
	}

	/*
	 * Border
	 */

	/**
	 * Sets the <i>border</i> property.
	 *  
	 * @param width The position value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border.asp">CSS border Property</a>
	 */

	public CSS setBorder(final Unit width, final BorderStyle style, final Serializable color) {
		return put("border", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border</i> property.
	 *  
	 * @param width The position value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border.asp">CSS border Property</a>
	 */

	public CSS setBorder(final BorderWidth width, final BorderStyle style, final Serializable color) {
		return put("border", width + " " + style + " " + color);
	}

	/**
	 *  Sets the <i>border-width</i> property.
	 *  
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-width.asp">CSS border-width Property</a>
	 */

	public CSS setBorderWidth(final Unit width) {
		return put("border-width", width);
	}

	/**
	 *  Sets the <i>border-width</i> property.
	 *  
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-width.asp">CSS border-width Property</a>
	 */

	public CSS setBorderWidth(final BorderWidth width) {
		return put("border-width", width);
	}

	/**
	 *  Sets the <i>border-style</i> property.
	 *  
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-style.asp">CSS border-style Property</a>
	 */

	public CSS setBorderStyle(final BorderStyle style) {
		return put("border-style", style);
	}

	/**
	 * Sets the <i>border-color</i> property.
	 *  
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-color.asp">CSS border-color Property</a>
	 */

	public CSS setBorderColor(final Serializable color) {
		return put("border-color", color);
	}

	// ----- TOP ----- 

	/**
	 * Sets the <i>border-top</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top.asp">CSS border-top Property</a>
	 */

	public CSS setBorderTop(final Unit width, final BorderStyle style, final Serializable color) {
		return put("border-top", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-top</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top.asp">CSS border-top Property</a>
	 */

	public CSS setBorderTop(final BorderWidth width, final BorderStyle style, final Serializable color) {
		return put("border-top", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-top-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top_width.asp">CSS border-top-width Property</a>
	 */

	public CSS setBorderTopWidth(final Unit width) {
		return put("border-top-width", width);
	}

	/**
	 * Sets the <i>border-top-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top_width.asp">CSS border-top-width Property</a>
	 */

	public CSS setBorderTopWidth(final BorderWidth width) {
		return put("border-top-width", width);
	}

	/**
	 * Sets the <i>border-top-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top_style.asp">CSS border-top-style Property</a>
	 */

	public CSS setBorderTopStyle(final BorderStyle style) {
		return put("border-top-style", style);
	}

	/**
	 * Sets the <i>border-top-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-top_color.asp">CSS border-top-color Property</a>
	 */

	public CSS setBorderTopColor(final Serializable color) {
		return put("border-top-color", color);
	}

	// ----- RIGHT ----- 

	/**
	 * Sets the <i>border-right</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right.asp">CSS border-right Property</a>
	 */

	public CSS setBorderRight(final Unit width, final BorderStyle style, final Serializable color) {
		return put("border-right", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-right</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right.asp">CSS border-right Property</a>
	 */

	public CSS setBorderRight(final BorderWidth width, final BorderStyle style, final Serializable color) {
		return put("border-right", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-right-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right_width.asp">CSS border-right-width Property</a>
	 */

	public CSS setBorderRightWidth(final Unit width) {
		return put("border-right-width", width);
	}

	/**
	 * Sets the <i>border-right-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right_width.asp">CSS border-right-width Property</a>
	 */

	public CSS setBorderRightWidth(final BorderWidth width) {
		return put("border-right-width", width);
	}

	/**
	 * Sets the <i>border-right-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right_style.asp">CSS border-right-style Property</a>
	 */

	public CSS setBorderRightStyle(final BorderStyle style) {
		return put("border-right-style", style);
	}

	/**
	 * Sets the <i>border-right-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-right_color.asp">CSS border-right-color Property</a>
	 */

	public CSS setBorderRightColor(final Serializable color) {
		return put("border-right-color", color);
	}

	// ----- LEFT ----- 

	/**
	 * Sets the <i>border-left</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left.asp">CSS border-left Property</a>
	 */

	public CSS setBorderLEft(final Unit width, final BorderStyle style, final Serializable color) {
		return put("border-left", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-left</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left.asp">CSS border-left Property</a>
	 */

	public CSS setBorderLeft(final BorderWidth width, final BorderStyle style, final Serializable color) {
		return put("border-left", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-left-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left_width.asp">CSS border-left-width Property</a>
	 */

	public CSS setBorderLeftWidth(final Unit width) {
		return put("border-left-width", width);
	}

	/**
	 * Sets the <i>border-left-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left_width.asp">CSS border-left-width Property</a>
	 */

	public CSS setBorderLeftWidth(final BorderWidth width) {
		return put("border-left-width", width);
	}

	/**
	 * Sets the <i>border-left-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left_style.asp">CSS border-left-style Property</a>
	 */

	public CSS setBorderLeftStyle(final BorderStyle style) {
		return put("border-left-style", style);
	}

	/**
	 * Sets the <i>border-left-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-left_color.asp">CSS border-left-color Property</a>
	 */

	public CSS setBorderLeftColor(final Serializable color) {
		return put("border-left-color", color);
	}

	// ----- BOTTOM ----- 

	/**
	 * Sets the <i>border-bottom</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom.asp">CSS border-bottom Property</a>
	 */

	public CSS setBorderBottom(final Unit width, final BorderStyle style, final Serializable color) {
		return put("border-bottom", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-bottom</i> property.
	 * 
	 * @param width The width value to be set.
	 * @param style The style value to be set.
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom.asp">CSS border-bottom Property</a>
	 */

	public CSS setBorderBottom(final BorderWidth width, final BorderStyle style, final Serializable color) {
		return put("border-bottom", width + " " + style + " " + color);
	}

	/**
	 * Sets the <i>border-bottom-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom_width.asp">CSS border-bottom-width Property</a>
	 */

	public CSS setBorderBottomWidth(final Unit width) {
		return put("border-bottom-width", width);
	}

	/**
	 * Sets the <i>border-bottom-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom_width.asp">CSS border-bottom-width Property</a>
	 */

	public CSS setBorderBottomWidth(final BorderWidth width) {
		return put("border-bottom-width", width);
	}

	/**
	 * Sets the <i>border-bottom-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom_style.asp">CSS border-bottom-style Property</a>
	 */

	public CSS setBorderBottomStyle(final BorderStyle style) {
		return put("border-bottom-style", style);
	}

	/**
	 * Sets the <i>border-bottom-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_border-bottom_color.asp">CSS border-bottom-color Property</a>
	 */

	public CSS setBorderBottomColor(final Serializable color) {
		return put("border-bottom-color", color);
	}

	/*
	 * Outline
	 */

	/**
	 * Sets the <i>outline</i> property.
	 * 
	 * @param color The color value to be set.
	 * @param style The style value to be set.
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline.asp">CSS outline Property</a>
	 */

	public CSS setOutline(final Serializable color, final OutlineStyle style, final Unit width) {
		return put("outline", color + " " + style + " " + width);
	}

	/**
	 * Sets the <i>outline</i> property.
	 * 
	 * @param color The color value to be set.
	 * @param style The style value to be set.
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline.asp">CSS outline Property</a>
	 */

	public CSS setOutline(final Serializable color, final OutlineStyle style, final OutlineWidth width) {
		return put("outline", color + " " + style + " " + width);
	}

	/**
	 * Sets the <i>outline-color</i> property.
	 * 
	 * @param color The color value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline-color.asp">CSS outline-color Property</a>
	 */

	public CSS setOutlineColor(final Serializable color) {
		return put("outline-color", color);
	}

	/**
	 * Sets the <i>outline-style</i> property.
	 * 
	 * @param style The style value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline-style.asp">CSS outline-style Property</a>
	 */

	public CSS setOutlineStyle(final OutlineStyle style) {
		return put("outline-style", style);
	}

	/**
	 * Sets the <i>outline-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline-width.asp">CSS outline-width Property</a>
	 */

	public CSS setOutlineWidth(final Unit width) {
		return put("outline-width", width);
	}

	/**
	 * Sets the <i>outline-width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_outline-width.asp">CSS outline-width Property</a>
	 */

	public CSS setOutlineStyle(final OutlineWidth width) {
		return put("outline-width", width);
	}

	/*
	 * Margin
	 */

	/**
	 * Sets the <i>margin</i> property.
	 *  
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin.asp">CSS margin Property</a>
	 */

	public CSS setMargin(final Unit margin) {
		return put("margin", margin);
	}

	/**
	 * Sets the <i>margin</i> property.
	 * 
	 * @param topBottom The top and bottom values to be set.
	 * @param rightLeft The right and left values to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin.asp">CSS margin Property</a>
	 */

	public CSS setMargin(final Unit topBottom, final Unit rightLeft) {
		return put("margin", topBottom + " " + rightLeft);
	}

	/**
	 * Sets the <i>margin</i> property.
	 * 
	 * @param top The top value to be set.
	 * @param rightLeft The right and left values to be set.
	 * @param bottom The bottom value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin.asp">CSS margin Property</a>
	 */

	public CSS setMargin(final Unit top, final Unit rightLeft, final Unit bottom) {
		return put("margin", top + " " + rightLeft + " " + bottom);
	}

	/**
	 * Sets the <i>margin</i> property.
	 * 
	 * @param top The top value to be set.
	 * @param right The right value to be set.
	 * @param bottom The bottom value to be set.
	 * @param left The left value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin.asp">CSS margin Property</a>
	 */

	public CSS setMargin(final Unit top, final Unit right, final Unit bottom, final Unit left) {
		return put("margin", top + " " + right + " " + bottom + " " + left);
	}

	/**
	 * Sets the <i>margin-bottom</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin-bottom.asp">CSS margin-bottom Property</a>
	 */

	public CSS setMarginBottom(final Unit margin) {
		return put("margin-bottom", margin);
	}

	/**
	 * Sets the <i>margin-left</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin-left.asp">CSS margin-left Property</a>
	 */

	public CSS setMarginLeft(final Unit margin) {
		return put("margin-left", margin);
	}

	/**
	 * Sets the <i>margin-right</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin-right.asp">CSS margin-right Property</a>
	 */

	public CSS setMarginRight(final Unit margin) {
		return put("margin-right", margin);
	}

	/**
	 * Sets the <i>margin-top</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_margin-top.asp">CSS margin-top Property</a>
	 */

	public CSS setMarginTop(final Unit margin) {
		return put("margin-top", margin);
	}

	/*
	 * Padding
	 */

	/**
	 * Sets the <i>padding</i> property.
	 * 
	 * @param padding The padding value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_padding.asp">CSS Padding</a>
	 */

	public CSS setPadding(final Unit padding) {
		return put("padding", padding);
	}

	/**
	 * Sets the <i>padding</i> property.
	 * 
	 * @param topBottom The top and bottom values to be set.
	 * @param rightLeft The right and left values to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_padding.asp">CSS Padding</a>
	 */

	public CSS setPadding(final Unit topBottom, final Unit rightLeft) {
		return put("padding", topBottom + " " + rightLeft);
	}

	/**
	 * Sets the <i>padding</i> property.
	 * 
	 * @param top The top value to be set.
	 * @param rightLeft The right and left values to be set.
	 * @param bottom The bottom value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_padding.asp">CSS Padding</a>
	 */

	public CSS setPadding(final Unit top, final Unit rightLeft, final Unit bottom) {
		return put("padding", top + " " + rightLeft + " " + bottom);
	}

	/**
	 * Sets the <i>padding</i> property.
	 * 
	 * @param top The top value to be set.
	 * @param right The right value to be set.
	 * @param bottom The bottom value to be set.
	 * @param left The left value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_padding.asp">CSS Padding</a>
	 */

	public CSS setPadding(final Unit top, final Unit right, final Unit bottom, final Unit left) {
		return put("padding", top + " " + right + " " + bottom + " " + left);
	}

	/**
	 * Sets the <i>padding-bottom</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_padding-bottom.asp">CSS padding-bottom</a>
	 */

	public CSS setPaddingBottom(final Unit margin) {
		return put("padding-bottom", margin);
	}

	/**
	 * Sets the <i>padding-left</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_padding-left.asp">CSS padding-left</a>
	 */

	public CSS setPaddingLeft(final Unit margin) {
		return put("padding-left", margin);
	}

	/**
	 * Sets the <i>padding-right</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_padding-right.asp">CSS padding-right</a>
	 */

	public CSS setPaddingRight(final Unit margin) {
		return put("padding-right", margin);
	}

	/**
	 * Sets the <i>padding-top</i> property.
	 * 
	 * @param margin The margin value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_padding-top.asp">CSS padding-top</a>
	 */

	public CSS setPaddingTop(final Unit margin) {
		return put("padding-top", margin);
	}

	/*
	 * Dimensions
	 */

	/**
	 * Sets the <i>width</i> property.
	 * 
	 * @param width The width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_width.asp">CSS width property</a>
	 */

	public CSS setWidth(final Unit width) {
		return put("width", width);
	}

	/**
	 * Sets the <i>min-width</i> property.
	 * 
	 * @param width The minimal width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_min-width.asp">CSS min-width property</a>
	 */

	public CSS setMinWidth(final Unit width) {
		return put("min-width", width);
	}

	/**
	 * Sets the <i>max-width</i> property.
	 * 
	 * @param width The maximum width value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_max-width.asp">CSS max-width property</a>
	 */

	public CSS setMaxWidth(final Unit width) {
		return put("max-width", width);
	}

	/**
	 * Sets the <i>height</i> property.
	 * 
	 * @param height The height value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_height.asp">CSS height property</a>
	 */

	public CSS setHeight(final Unit height) {
		return put("height", height);
	}

	/**
	 * Sets the <i>min-height</i> property.
	 * 
	 * @param height The height value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_min-height.asp">CSS min-height property</a>
	 */

	public CSS setMinHeight(final Unit height) {
		return put("min-height", height);
	}

	/**
	 * Sets the <i>max-height</i> property.
	 * 
	 * @param height The height value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_dim_max-height.asp">CSS max-height property</a>
	 */

	public CSS setMaxHeight(final Unit height) {
		return put("max-height", height);
	}

	/*
	 * Position
	 */

	/**
	 * Sets the <i>bottom</i> property.
	 * 
	 * @param bottom The bottom value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_bottom.asp">CSS bottom Property</a>
	 */

	public CSS setBottom(final Unit bottom) {
		return put("bottom", bottom);
	}

	/**
	 * Sets the <i>left</i> property.
	 * 
	 * @param left The left value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_left.asp">CSS left Property</a>
	 */

	public CSS setLeft(final Unit left) {
		return put("left", left);
	}

	/**
	 * Sets the <i>right</i> property.
	 * 
	 * @param right The right value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_right.asp">CSS right Property</a>
	 */

	public CSS setRight(final Unit right) {
		return put("right", right);
	}

	/**
	 * Sets the <i>top</i> property.
	 * 
	 * @param top The top value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_top.asp">CSS top Property</a>
	 */

	public CSS setTop(final Unit top) {
		return put("top", top);
	}

	/**
	 * Sets the <i>clip</i> property.
	 * 
	 * @param top The top value to be set.
	 * @param right The right value to be set.
	 * @param bottom The bottom value to be set.
	 * @param left The left value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_clip.asp">CSS clip Property</a>
	 */

	public CSS setClipShape(final Unit top, final Unit right, final Unit bottom, final Unit left) {
		return put("clip", "rect(" + top + ", " + right + ", " + bottom + ", " + left + ")");
	}

	/**
	 * Sets the <i>cursor</i> property.
	 * 
	 * @param cursor The cursor value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_cursor.asp">CSS cursor Property</a>
	 */

	public CSS setCursor(final Cursor cursor) {
		return put("cursor", cursor);
	}

	/**
	 * Sets the <i>cursor</i> property.
	 * 
	 * @param url The cursor url value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_cursor.asp">CSS cursor Property</a>
	 */

	public CSS setCursor(final Serializable url) {
		return put("cursor", "url(" + url + ")");
	}

	/**
	 * Sets the <i>overflow</i> property.
	 * 
	 * @param overflow The overflow value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_overflow.asp">CSS overflow Property</a>
	 */

	public CSS setOverflow(final Overflow overflow) {
		return put("overflow", overflow);
	}

	/**
	 * Sets the <i>position</i> property.
	 * 
	 * @param position The position value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/css/css_positioning.asp">CSS Positioning</a>
	 */

	public CSS setPosition(final Position position) {
		return put("position", position);
	}

	/**
	 * Sets the <i>z-index</i> property.
	 * 
	 * @param zIndex The z-index value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_pos_z-index.asp">CSS z-index Property</a>
	 */

	public CSS setZIndex(final int zIndex) {
		return put("z-index", zIndex);
	}

	/*
	 * Float
	 */

	/**
	 * Sets the <i>clear</i> property.
	 * 
	 * @param clear The clear value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_clear.asp">CSS clear Property</a>
	 */

	public CSS setClear(final Clear clear) {
		return put("clear", clear);
	}

	/**
	 * Sets the <i>float</i> property.
	 * 
	 * @param f The float value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_float.asp">CSS float Property</a>
	 */

	public CSS setFloat(final Float f) {
		return put("float", f);
	}

	/*
	 * Display
	 */

	/**
	 * Sets the <i>display</i> property.
	 * 
	 * @param display The display value to be set.
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_display.asp">CSS display Property</a>
	 */

	public CSS setDisplay(final Display display) {
		return put("display", display);
	}

	/**
	 * Sets the <i>visibility</i> property.
	 * 
	 * @param visibility The visibility to be set
	 * @return The current CSS object.
	 * @see <a href="http://www.w3schools.com/cssref/pr_class_visibility.asp">CSS visibility Property</a>
	 */

	public CSS setVisibility(final Visibility visibility) {
		return put("visibility", visibility);
	}

}