/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 */

package com.osbcp.css;

import java.io.File;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

/**
 * Exporter class
 * 
 * @author <a href="mailto:christoffer@christoffer.me">Christoffer Pettersson</a>
 */

final class Exporter {

	/**
	 * Exports to CSS to a file.
	 * 
	 * @param stylesheetClass The class where the CSS is defined .
	 * @param stylesheetFile The file the CSS should be written to.
	 * @param includeComments True if comments should be written as well.
	 * @throws Exception If any error occurs
	 */

	final static void export(final Class<?> stylesheetClass, final File stylesheetFile, final boolean includeComments) throws Exception {

		// Create the file writer for files
		PrintStream out = new PrintStream(stylesheetFile);

		System.out.println("Investigating '" + stylesheetClass.getName() + "'.");

		// For each declared field in the class
		for (final Field field : stylesheetClass.getDeclaredFields()) {

			// Check that the field is eligible
			if (!isFieldEligible(field)) {
				continue;
			}

			// Get the object
			CSS css = (CSS) field.get(null);

			// Check if the value is null
			if (css == null) {
				System.out.println("Field '" + field.getName() + "' is not eligible. Value is null.");
				continue;
			}

			// Write debug message
			System.out.println("Found eligible field '" + field.getName() + "'.");

			writeCSStoFile(out, css, stylesheetClass, field, includeComments);

		}

		/*
		 * Print out the done message
		 */

		System.out.println("CSS exported to '" + stylesheetFile.getAbsolutePath() + "'.");

		// Close the writer
		out.close();

	}

	/**
	 * Writes a structured CSS to the file
	 * 
	 * @param out The file's print stream
	 * @param css CSS to write
	 * @param stylesheetClass The class where the CSS is defined 
	 * @param field The field where the CSS is defined
	 * @param includeComments True if comments should be written as well. 
	 * @throws Exception If any error occurs
	 */

	private static void writeCSStoFile(final PrintStream out, final CSS css, final Class<?> stylesheetClass, final Field field, final boolean includeComments) throws Exception {

		// Loop through all the JavaScript objects
		if (includeComments) {
			out.println("/*");
			out.println(" * " + stylesheetClass.getSimpleName() + "." + field.getName());
			out.println(" */\n");
		}

		out.print(css.toStringBlock());

	}

	/**
	 * Checks if a field is eligible to be exported
	 * 
	 * @param field The field to check
	 * @return True if the field is eligible
	 */

	private static boolean isFieldEligible(final Field field) {

		// Check if the type is CSS
		if (field.getType() != CSS.class) {
			System.out.println("Field '" + field.getName() + "' is not eligible. Type is '" + field.getType() + "' and not '" + CSS.class.getCanonicalName() + "'.");
			return false;
		}

		// Check if the field is static
		if (!Modifier.isStatic(field.getModifiers())) {
			System.out.println("Field '" + field.getName() + "' is not eligible. Field is not static.");
			return false;
		}

		// Field is eligible
		return true;

	}

}